/*
 * @(#)Tertris.c
 *
 * Copyright 1993 - 2010  David A. Bagley, bagleyd@tux.org
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao@cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Methods file for Tertris */

#include "file.h"
#include "rngs.h"
#include "sound.h"
#include "TertrisP.h"

static Point triangle_unit[MAX_ORIENT][3] = {
	{ {0, 0}, {-1, -1}, { 2, 0} },
	{ {0, 0}, { 1, 1}, {-2, 0} }
};
static Point triangle_list[MAX_ORIENT][4];
static Point triangle_border[MAX_ORIENT][4];

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wtertris.ini"
#endif

#define SECTION "setup"
#else

#if defined( USE_SOUND ) && defined( USE_NAS )
Display *dsp;
#endif

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void resizeGame(TertrisWidget w);
static void sizeGame(TertrisWidget w);
static Boolean setValuesGame(Widget current, Widget request, Widget renew);
static void destroyGame(Widget old);
static void quitGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void initializeGame(Widget request, Widget renew);
static void exposeGame(Widget renew, XEvent *event, Region region);
static void mouseGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void hideGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void clearGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void enterGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void leaveGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void reflectGame(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void rotateGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void moveGameLeft(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void moveGameRight(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void dropGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void moveGameInput(TertrisWidget w, int direction);
static void incrementGameLevel(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameLevel(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void incrementGameUnits(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameUnits(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void incrementGameColumns(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameColumns(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void fixedfillGame(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void suddenAppearGame(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void diagonalGame(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void mixedGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void bonusGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void cwGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void plainPieceGame(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void showNextGame(TertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void gridGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void soundGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void pauseGame(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill0Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill1Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill2Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill3Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill4Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill5Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill6Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill7Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill8Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill9Game(TertrisWidget w, XEvent *event, char **args, int nArgs);
static void fill10Game(TertrisWidget w, XEvent *event, char **args, int nArgs);

static char defaultTranslationsGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <Btn4Down>: Mouse() Reflect()\n\
 <KeyPress>i: Reflect()\n\
 <KeyPress>e: Reflect()\n\
 <KeyPress>F11: Reflect()\n\
 <KeyPress>KP_Divide: Reflect()\n\
 <KeyPress>R5: Reflect()\n\
 <Btn2Down>: Mouse() Rotate()\n\
 <KeyPress>k: Rotate()\n\
 <KeyPress>d: Rotate()\n\
 <KeyPress>Up: Rotate()\n\
 <KeyPress>osfUp: Rotate()\n\
 <KeyPress>KP_Up: Rotate()\n\
 <KeyPress>R8: Rotate()\n\
 <KeyPress>KP_8: Rotate()\n\
 <KeyPress>F12: Rotate()\n\
 <KeyPress>Begin: Rotate()\n\
 <KeyPress>KP_5: Rotate()\n\
 <KeyPress>R11: Rotate()\n\
 <Btn1Down>: Mouse() MoveLeft()\n\
 <KeyPress>j: MoveLeft()\n\
 <KeyPress>s: MoveLeft()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <Btn3Down>: Mouse() MoveRight()\n\
 <KeyPress>l: MoveRight()\n\
 <KeyPress>f: MoveRight()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <Btn5Down>: Mouse() Drop()\n\
 <KeyPress>Return: Drop()\n\
 <KeyPress>KP_Enter: Drop()\n\
 <KeyPress>Down: Drop()\n\
 <KeyPress>osfDown: Drop()\n\
 <KeyPress>KP_Down: Drop()\n\
 <KeyPress>KP_2: Drop()\n\
 <KeyPress>R14: Drop()\n\
 <KeyPress>KP_Space: Drop()\n\
 <KeyPress>space: Drop()\n\
 <Btn4Down>: Mouse() Rotate()\n\
 <Btn5Down>: Mouse() Drop()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>>: IncrementLevel()\n\
 <KeyPress>.: IncrementLevel()\n\
 <KeyPress><: DecrementLevel()\n\
 <KeyPress>+: IncrementUnits()\n\
 <KeyPress>=: IncrementUnits()\n\
 <KeyPress>-: DecrementUnits()\n\
 <KeyPress>_: DecrementUnits()\n\
 <KeyPress>]: IncrementColumns()\n\
 <KeyPress>}: IncrementColumns()\n\
 <KeyPress>[: DecrementColumns()\n\
 <KeyPress>{: DecrementColumns()\n\
 <KeyPress>x: FixedFill()\n\
 <KeyPress>r: SuddenAppear()\n\
 <KeyPress>g: Diagonal()\n\
 <KeyPress>m: Mixed()\n\
 <KeyPress>b: Bonus()\n\
 <KeyPress>w: Cw()\n\
 <KeyPress>p: PlainPiece()\n\
 <KeyPress>n: ShowNext()\n\
 Shift<KeyPress>3: Grid()\n\
 Shift<KeyPress>2: Sound()\n\
 <KeyPress>|: Pause()\n\
 <KeyPress>`: Fill0()\n\
 <KeyPress>1: Fill1()\n\
 <KeyPress>2: Fill2()\n\
 <KeyPress>3: Fill3()\n\
 <KeyPress>4: Fill4()\n\
 <KeyPress>5: Fill5()\n\
 <KeyPress>6: Fill6()\n\
 <KeyPress>7: Fill7()\n\
 <KeyPress>8: Fill8()\n\
 <KeyPress>9: Fill9()\n\
 <KeyPress>0: Fill10()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

/*
 <KeyPress>,: DecrementLevel()\n\
#if defined(UP_REFLECT)
 <KeyPress>KP_Up: Reflect()\n\
 <KeyPress>R8: Reflect()\n\
 <KeyPress>KP_8: Reflect()\n\
#endif
*/

static XtActionsRec actionsListGame[] =
{
	{(char *) "Quit", (XtActionProc) quitGame},
	{(char *) "Hide", (XtActionProc) hideGame},
	{(char *) "Mouse", (XtActionProc) mouseGame},
	{(char *) "Reflect", (XtActionProc) reflectGame},
	{(char *) "Rotate", (XtActionProc) rotateGame},
	{(char *) "MoveLeft", (XtActionProc) moveGameLeft},
	{(char *) "MoveRight", (XtActionProc) moveGameRight},
	{(char *) "Drop", (XtActionProc) dropGame},
	{(char *) "Clear", (XtActionProc) clearGame},
	{(char *) "IncrementLevel", (XtActionProc) incrementGameLevel},
	{(char *) "DecrementLevel", (XtActionProc) decrementGameLevel},
	{(char *) "IncrementUnits", (XtActionProc) incrementGameUnits},
	{(char *) "DecrementUnits", (XtActionProc) decrementGameUnits},
	{(char *) "IncrementColumns", (XtActionProc) incrementGameColumns},
	{(char *) "DecrementColumns", (XtActionProc) decrementGameColumns},
	{(char *) "FixedFill", (XtActionProc) fixedfillGame},
	{(char *) "SuddenAppear", (XtActionProc) suddenAppearGame},
	{(char *) "Diagonal", (XtActionProc) diagonalGame},
	{(char *) "Mixed", (XtActionProc) mixedGame},
	{(char *) "Bonus", (XtActionProc) bonusGame},
	{(char *) "Cw", (XtActionProc) cwGame},
	{(char *) "PlainPiece", (XtActionProc) plainPieceGame},
	{(char *) "ShowNext", (XtActionProc) showNextGame},
	{(char *) "Grid", (XtActionProc) gridGame},
	{(char *) "Sound", (XtActionProc) soundGame},
	{(char *) "Pause", (XtActionProc) pauseGame},
	{(char *) "Fill0", (XtActionProc) fill0Game},
	{(char *) "Fill1", (XtActionProc) fill1Game},
	{(char *) "Fill2", (XtActionProc) fill2Game},
	{(char *) "Fill3", (XtActionProc) fill3Game},
	{(char *) "Fill4", (XtActionProc) fill4Game},
	{(char *) "Fill5", (XtActionProc) fill5Game},
	{(char *) "Fill6", (XtActionProc) fill6Game},
	{(char *) "Fill7", (XtActionProc) fill7Game},
	{(char *) "Fill8", (XtActionProc) fill8Game},
	{(char *) "Fill9", (XtActionProc) fill9Game},
	{(char *) "Fill10", (XtActionProc) fill10Game},
	{(char *) "Enter", (XtActionProc) enterGame},
	{(char *) "Leave", (XtActionProc) leaveGame}
};

static XtResource resourcesGame[] =
{
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(TertrisWidget, core.width),
	 XtRString, (caddr_t) "281"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(TertrisWidget, core.height),
	 XtRString, (caddr_t) "609"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(TertrisWidget, tertris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(TertrisWidget, tertris.background),
	 XtRString, (caddr_t) "#AEB2C3" /*XtDefaultBackground*/},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.delay),
	 XtRString, (caddr_t) "1000"},
	{XtNsound, XtCSound, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.sound),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbumpSound, XtCBumpSound, XtRString, sizeof (String),
	 XtOffset(TertrisWidget, tertris.bumpSound),
	 XtRString, (caddr_t) BUMPSOUND},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.columns),
	 XtRString, (caddr_t) "15"},
	{XtNlevel, XtCLevel, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.level),
	 XtRString, (caddr_t) "3"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.units),
	 XtRString, (caddr_t) "5"},
	{XtNfill, XtCFill, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.fill),
	 XtRString, (caddr_t) "0"},
	{XtNfixedFill, XtCFixedFill, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.fixedFill),
	 XtRString, (caddr_t) "TRUE"},
	{XtNsuddenAppear, XtCSuddenAppear, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.suddenAppear),
	 XtRString, (caddr_t) "FALSE"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.bonus),
	 XtRString, (caddr_t) "TRUE"},
	{XtNcw, XtCCw, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.cw),
	 XtRString, (caddr_t) "TRUE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNgrid, XtCGrid, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.grid),
	 XtRString, (caddr_t) "TRUE"},
	{XtNscore, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.score),
	 XtRString, (caddr_t) "0"},
	{XtNcompletedRows, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.completedRows),
	 XtRString, (caddr_t) "0"},
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(TertrisWidget, tertris.userName),
	 XtRString, (caddr_t) ""},
	{XtNscoreFile, XtCScoreFile, XtRString, sizeof (String),
	 XtOffset(TertrisWidget, tertris.scoreFile),
	 XtRString, (caddr_t) ""},
	{XtNscoreOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.scoreOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNversionOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.versionOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNact, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TertrisWidget, tertris.act),
	 XtRString, (caddr_t) "FALSE"},
	{XtNpixmapSize, XtCPixmapSize, XtRInt, sizeof (int),
	 XtOffset(TertrisWidget, tertris.pixmapSize),
	 XtRString, (caddr_t) "64"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(TertrisWidget, tertris.select),
	 XtRCallback, (caddr_t) NULL}
};

TertrisClassRec tertrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Tertris",	/* class name */
		sizeof (TertrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) initializeGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListGame,	/* actions */
		XtNumber(actionsListGame),	/* num actions */
		resourcesGame,	/* resources */
		XtNumber(resourcesGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) destroyGame,	/* destroy */
		(XtWidgetProc) resizeGame,	/* resize */
		(XtExposeProc) exposeGame,	/* expose */
		(XtSetValuesFunc) setValuesGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass tertrisWidgetClass = (WidgetClass) & tertrisClassRec;

static void
setGame(TertrisWidget w, int reason)
{
	tertrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

static void
setGameNext(TertrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	tertrisCallbackStruct cb;

	cb.reason = reason;
	cb.randomNumber = randomNumber;
	cb.randomRotation = randomRotation;
	cb.randomReflection = randomReflection;
	cb.units = units;
	cb.diagonal = diagonal;
	cb.mixed = mixed;
	cb.newBonus = newBonus;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static const int delay[MAX_LEVELS] = {
	100, 92, 84, 75, 68, 60, 52, 45, 39, 32,
	26, 21, 16, 12, 8, 5, 3, 2, 1, 0
};

static void
checkThings(TertrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->tertris.level < MIN_LEVELS) {
		w->tertris.level = MIN_LEVELS;
		intCat(&buf1, "Number of levels out of bounds, use at least ",
			MIN_LEVELS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.level);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	} else if (w->tertris.level > MAX_LEVELS) {
		w->tertris.level = MAX_LEVELS;
		intCat(&buf1, "Number of levels out of bounds, use at most ",
			MAX_LEVELS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.level);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->tertris.units < MIN_UNITS) {
		w->tertris.units = MIN_UNITS;
		w->tertris.next.units = w->tertris.units;
		intCat(&buf1, "Number of units out of bounds, use at least ",
			MIN_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	} else if (w->tertris.units > MAX_UNITS) {
		w->tertris.units = MAX_UNITS;
		w->tertris.next.units = w->tertris.units;
		intCat(&buf1, "Number of units out of bounds, use at most ",
			MAX_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->tertris.columns < MIN_COLUMNS) {
		w->tertris.columns = MIN_COLUMNS;
		intCat(&buf1, "Number of columns out of bounds, use at least ",
			MIN_COLUMNS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.columns);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->tertris.columns % 2 == 0) {
		w->tertris.columns--;
		intCat(&buf1,
			"Number of columns must be odd, defaulting to ",
			w->tertris.columns);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (!LEGAL(w->tertris.units, w->tertris.diagonal,
			w->tertris.mixed, FALSE)) {
		intCat(&buf1, "Illegal configuration of units ",
			w->tertris.units);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->tertris.diagonal);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->tertris.mixed);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to units ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULT_DIAGONAL);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULT_MIXED);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->tertris.units = DEFAULT_UNITS;
		w->tertris.diagonal = FALSE;
		w->tertris.mixed = FALSE;
		w->tertris.next.units = w->tertris.units;
		w->tertris.next.diagonal = w->tertris.diagonal;
		w->tertris.next.mixed = w->tertris.mixed;
	}
	if (w->tertris.fill < 0) {
		w->tertris.fill = 0;
		DISPLAY_WARNING("Number for fill can not be negative, defaulting to 0");
	} else if (w->tertris.fill > MAX_FILL) {
		w->tertris.fill = MAX_FILL;
		intCat(&buf1, "Number for fill out of bounds, use at most ",
			MAX_FILL);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.fill);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
}

#define BRIGHT_FACTOR 0.8
#define DARK_FACTOR 0.75

#ifdef WINVER
#define MAX_INTENSITY 0xFF
int
brighter(const int light)
{
	int i = (int) ((1 - BRIGHT_FACTOR) * MAX_INTENSITY);
	int temp = light;

	if (temp < i)
		temp = i;
	return MIN(temp / BRIGHT_FACTOR, MAX_INTENSITY);
}

int
darker(const int light)
{
	return (int) (light * DARK_FACTOR);
}

#else
#define MAX_INTENSITY 0xFFFF

Pixel
brighter(TertrisWidget w, Pixel pixel)
{
	XColor color;
	Colormap colormap = DefaultColormapOfScreen(XtScreen(w));
	int i = (int) ((1 - BRIGHT_FACTOR) * MAX_INTENSITY);

	color.pixel = pixel;
	VOID XQueryColor(XtDisplay(w), colormap, &color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / BRIGHT_FACTOR, MAX_INTENSITY);
	color.green = (unsigned short) MIN(color.green / BRIGHT_FACTOR, MAX_INTENSITY);
	color.blue = (unsigned short) MIN(color.blue / BRIGHT_FACTOR, MAX_INTENSITY);
	if (XAllocColor(XtDisplay(w), colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Tertris: Could not allocate brighter color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}

Pixel
darker(TertrisWidget w, Pixel pixel)
{
	XColor color;
	Colormap colormap = DefaultColormapOfScreen(XtScreen(w));

	color.pixel = pixel;
	VOID XQueryColor(XtDisplay(w), colormap, &color);
	color.red = (unsigned short) (color.red * DARK_FACTOR);
	color.green = (unsigned short) (color.green * DARK_FACTOR);
	color.blue = (unsigned short) (color.blue * DARK_FACTOR);
	if (XAllocColor(XtDisplay(w), colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Tertris: Could not allocate darker color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}
#endif

#if 0
static void
drawTri(TertrisWidget w, Window win, GC gc, int i, int j)
{
	int orient = (i + j + !w->tertris.invert) % MAX_ORIENT;

	triangle_list[orient][0].x = (i + 1) * (w->tertris.triWidth / 2) + 1;
	triangle_list[orient][0].y = j * w->tertris.triHeight + DELTAY +
			((orient != 0) ? 0 : w->tertris.triHeight - DELTAY);
	VOID XFillPolygon(display, win, gc, triangle_list[orient], 3,
			Convex, CoordModePrevious);
}
#endif

static void
drawUnitBorder(TertrisWidget w, int i, int j)
{
	int orient = (i + j + !w->tertris.invert) % MAX_ORIENT;
	GC gc;

	if (w->tertris.grid)
		gc = w->tertris.inverseGC[3];
	else
		gc = w->tertris.inverseGC[2];
	triangle_border[orient][0].x =
		(i + 1) * (w->tertris.triWidth / 2) + w->tertris.offset.x;
	triangle_border[orient][0].y = j * w->tertris.triHeight +
		w->tertris.offset.y +
		((orient != 0) ? 0 : w->tertris.triHeight);
	POLYLINE(w, dr, gc, triangle_border[orient], 4, False);
}

#define EDGE 2
#define SQRT3_2 0.8660254 /* sqrt(3) / 2 */

void
fillPolygon(TertrisWidget w, int pm, int color, Boolean invert,
		Point *points, Boolean graphicalXor)
{
	int midPoint;
	Point tempList[4];
	GC gc[FG_SHADES];
	int i;

	for (i = 0; i < FG_SHADES; i++)
		gc[i] = w->tertris.thingGC[color][i];
	midPoint = (points[1].y + points[2].y) / 3;
#ifdef WINVER
	/* simulate Xor */
	if (graphicalXor) {
		for (i = 0; i < FG_SHADES; i++)
			gc[i] = RGB(255 - GetRValue(gc[i]),
				255 - GetGValue(gc[i]),
				255 - GetBValue(gc[i]));
	}
#endif
#if 0
	tempList[0].x = points[0].x;
	tempList[0].y = points[0].y;
	tempList[1].x = points[1].x;
	tempList[1].y = points[1].y;
	tempList[2].x = -points[1].x;
	tempList[2].y = -midPoint;
#endif
	POLYGON((TertrisWidget) w, dr, gc[1], gc[1],
		points, 3, True, False);
	if (!w->tertris.plainPiece) {
		Point insideList[4], border[5];
		int offset;

		tempList[0].x = points[0].x;
		tempList[0].y = points[0].y;
		for (i = 1; i < 4; i++) {
			tempList[i].x = tempList[i - 1].x + points[i].x;
			tempList[i].y = tempList[i - 1].y + points[i].y;
		}
		offset = (int) (midPoint / (SQRT3_2 * EDGE));
		for (i = 0; i < 4; i++) {
			insideList[i].x = ((EDGE - 1) * tempList[i].x +
				points[0].x) / EDGE;
			insideList[i].y = ((EDGE - 1) * tempList[i].y +
				points[0].y +
				midPoint) / EDGE + offset;
		}
		for (i = 0; i < 3; i++) {
			if ((!invert && ((pm & ((1 << (2 - ((i + 2) % 3))))) != 0)) ||
					(invert && ((pm & ((1 << (2 - ((i + 1) % 3))))) != 0))) {
				border[0] = tempList[i];
				border[1] = insideList[i];
				border[2] = insideList[(i + 1) % 3];
				border[3] = tempList[(i + 1) % 3];
				border[4] = tempList[i];
				POLYGON2((TertrisWidget) w, dr,
					gc[2], gc[0],
					border, 4, True, True);
			}
		}
		/*if ((!invert && (pm & 4)) || (invert && (pm & 1))) {
			border[0] = tempList[1];
			border[1] = insideList[1];
			border[2] = insideList[2];
			border[3] = tempList[2];
			border[4] = tempList[1];
			POLYGON((TertrisWidget) w, dr, gc[2], gc[0],
				border, 4, True, True);
		}
		if ((!invert && (pm & 2)) || (invert && (pm & 4))) {
			border[0] = tempList[2];
			border[1] = insideList[2];
			border[2] = insideList[0];
			border[3] = tempList[0];
			border[4] = tempList[2];
			POLYGON((TertrisWidget) w, dr, gc[2], gc[0],
				border, 4, True, True);
		}*/
		return;
	}
	tempList[0].x = points[0].x + points[1].x;
	tempList[0].y = points[0].y + points[1].y;
	tempList[1].x = points[2].x;
	tempList[1].y = points[2].y;
	tempList[2].x = points[1].x;
	tempList[2].y = -midPoint;
	tempList[3].x = points[1].x;
	tempList[3].y = midPoint;
	POLYGON((TertrisWidget) w, dr,
		((invert) ? gc[2] : gc[0]),
		((invert) ? gc[2] : gc[0]),
		tempList, 3, True, False);
	tempList[0].x = points[0].x;
	tempList[0].y = points[0].y;
	tempList[1].x = -points[1].x;
	tempList[1].y = points[1].y;
	tempList[2].x = points[1].x;
	tempList[2].y = -midPoint;
	tempList[3].x = 0;
	tempList[3].y = midPoint - points[1].y;
	POLYGON((TertrisWidget) w, dr,
		((invert) ? gc[0] : gc[2]),
		((invert) ? gc[0] : gc[2]),
		tempList, 3, True, False);
}

static void
drawUnit(TertrisWidget w, int pmid, int cid, int i, int j,
		Boolean graphicalXor)
{
	if (i >= 0 && i < w->tertris.columns &&
			j >= 0 && j < w->tertris.rows) {
		int orient = (i + j + !w->tertris.invert) % MAX_ORIENT;

		triangle_border[orient][0].x =
			(i + 1) * (w->tertris.triWidth / 2) +
			w->tertris.offset.x;
		triangle_border[orient][0].y = j * w->tertris.triHeight +
			w->tertris.offset.y +
			((orient != 0) ? 0 : w->tertris.triHeight);
		fillPolygon(w, pmid, cid, orient, triangle_border[orient],
			graphicalXor);
	}
}

static void
clearUnit(TertrisWidget w, int i, int j)
{
	if (i >= 0 && i < w->tertris.columns &&
			j >= 0 && j < w->tertris.rows) {
		int orient = (i + j + !w->tertris.invert) % MAX_ORIENT;

		triangle_border[orient][0].x =
			(i + 1) * (w->tertris.triWidth / 2) +
			w->tertris.offset.x;
		triangle_border[orient][0].y = j * w->tertris.triHeight +
			w->tertris.offset.y +
			((orient != 0) ? 0 : w->tertris.triHeight);
		POLYGON(w, dr, w->tertris.inverseGC[2], w->tertris.inverseGC[2],
			triangle_border[orient], 3,
			True, False);
	}
}

static void
xorUnit(TertrisWidget w, int i, int j, int k)
{
	int orient = (i + j + !w->tertris.invert) % MAX_ORIENT;

	triangle_border[orient][0].x =
		(i + 1) * (w->tertris.triWidth / 2) + w->tertris.offset.x;
	triangle_border[orient][0].y = j * w->tertris.triHeight +
		w->tertris.offset.y +
		((orient != 0) ? 0 : w->tertris.triHeight);
#ifdef WINVER
	if (w->tertris.field[j * w->tertris.columns + i].pmid >= 0) {
		drawUnit(w,
			w->tertris.field[j * w->tertris.columns + i].pmid,
			w->tertris.field[j * w->tertris.columns + i].cid,
			i, j, (k % 2 == 1));
	}
#else
	POLYGON(w, dr, w->tertris.xorGC, w->tertris.xorGC,
		triangle_border[orient], 3,
		True, False);
#endif
}

static void
moveUnit(TertrisWidget w, int i, int j)
{
	if (w->tertris.field[j * w->tertris.columns + i].pmid >= 0) {
		drawUnit(w,
			w->tertris.field[j * w->tertris.columns + i].pmid,
			w->tertris.field[j * w->tertris.columns + i].cid,
			i, j, False);
	} else {
		clearUnit(w, i, j);
	}
	drawUnitBorder(w, i, j);
}

static void
drawPartialField(TertrisWidget w, int start, int finish)
{
	int i, j, offset;

	for (j = start; j < finish; j++) {
		offset = j * w->tertris.columns;
		for (i = 0; i < w->tertris.columns; i++) {
			if (w->tertris.field[offset + i].pmid >= 0) {
				drawUnit(w, w->tertris.field[offset + i].pmid,
					w->tertris.field[offset + i].cid,
					i, j, False);
				drawUnitBorder(w, i, j);
			}
		}
	}
}

static void
drawField(TertrisWidget w)
{
	drawPartialField(w, 0, w->tertris.rows);
}

#if 0
static void
printField(TertrisWidget w)
{
	int i, j, offset;

	for (j = 0; j < w->tertris.rows; j++) {
		offset = j * w->tertris.columns;
		for (i = 0; i < w->tertris.columns; i++) {
			if (w->tertris.field[offset + i].pmid >= 0)
				(void) printf("*");
			else
				(void) printf(" ");
		}
		(void) printf("\n");
	}
}
#endif

static void
clearField(TertrisWidget w)
{
	int i, j;

	for (j = 0; j < w->tertris.rows; j++) {
		for (i = 0; i < w->tertris.columns; i++) {
			clearUnit(w, i, j);
		}
	}
}

static void
drawThing(TertrisWidget w)
{
	int i, j;

	for (i = 0; i < HEX_COL(w->tertris.curThing.size); i++) {
		int xi = w->tertris.curThing.xpos + i;

		for (j = 0; j < HEX_ROW(w->tertris.curThing.size); j++) {
			int yj = w->tertris.curThing.ypos + j;

			if (yj >= 0 && inHexagon(w->tertris.curThing.size, i, j) != 0 &&
					tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape), w->tertris.curThing.size),
					w->tertris.curThing.colorNumber,
					xi, yj, False);
				drawUnitBorder(w, xi, yj);
			}
		}
	}
}

static void
drawThingDiff(TertrisWidget w, Thing oldThing)
{
	int i, j;
	int sx = HEX_COL(w->tertris.curThing.size);
	int sy = HEX_ROW(w->tertris.curThing.size);

	for (i = 0; i < sx; i++) {
		int xi = w->tertris.curThing.xpos + i;

		for (j = 0; j < sy; j++) {
			int yj = w->tertris.curThing.ypos + j;

			if (yj >= 0 && inHexagon(w->tertris.curThing.size, i, j) != 0 &&
					tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape), w->tertris.curThing.size),
					w->tertris.curThing.colorNumber,
					xi, yj, False);
				drawUnitBorder(w, xi, yj);
			}
		}
	}

	for (i = 0; i < sx; i++) {
		int xi = oldThing.xpos + i;
		int ox = xi - w->tertris.curThing.xpos;

		if (xi >= 0 && xi < w->tertris.columns)
			for (j = 0; j < sy; j++) {

				int yj = oldThing.ypos + j;
				int oy = yj - w->tertris.curThing.ypos;

				if (yj >= 0 && yj < w->tertris.rows &&
					tris[oldThing.units].polyiamond
					[oldThing.polyNumber].shape[j][i] != 0 &&
					((ox < 0) || (ox >= sx) ||
					(oy < 0) || (oy >= sy) ||
					tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape[oy][ox] == 0)) {
					clearUnit(w, xi, yj);
					drawUnitBorder(w, xi, yj);
				}
			}
	}
}

static void
drawGrid(TertrisWidget w)
{
	int i, j;

	for (j = 0; j < w->tertris.rows; j++)
		for (i = 0; i < w->tertris.columns; i++)
			drawUnitBorder(w, i, j);
}

static void
draw3DFrame(const TertrisWidget w, Boolean inside,
	int x, int y, int width, int height)
{
	GC gc;

	if (inside)
		gc = w->tertris.inverseGC[3];
	else
		gc = w->tertris.inverseGC[2];
	FILLRECTANGLE(w, dr, gc,
		x, y, width, 1);
	FILLRECTANGLE(w, dr, gc,
		x, y + 1, 1, height - 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 1, width - 2, 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 2, 1, height - 3);
	if (inside)
		gc = w->tertris.inverseGC[0];
	else
		gc = w->tertris.inverseGC[2];
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + height - 1, width - 1, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 1, y + 1, 1, height - 2);
	FILLRECTANGLE(w, dr, gc,
		x + 2, y + height - 2, width - 3, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 2, y + 2, 1, height - 4);
}

static void
eraseFrame(TertrisWidget w)
{
	FILLRECTANGLE(w, dr, w->tertris.inverseGC[2],
		0, 0, w->core.width, w->core.height);
}

static void
drawFrame(TertrisWidget w, Boolean focus)
{
	int offsetX = w->tertris.offset.x - 2;
	int offsetY = w->tertris.offset.y - 2;
	int lengthX = w->tertris.offset.x + w->tertris.viewWidth + 2;
	int lengthY = w->tertris.offset.y + w->tertris.viewHeight + 2;

	FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
		0, 0, w->core.width, offsetY);
	FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
		0, offsetY, offsetX, lengthY - offsetY);
	FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
		lengthX, offsetY, w->core.width - lengthX, lengthY - offsetY);
	FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
		0, lengthY, w->core.width, offsetY + 1);
	draw3DFrame(w, focus,
		offsetX, offsetY, lengthX - offsetX, lengthY - offsetY);
}

static void
drawNext(TertrisWidget w)
{
	setGameNext(w, ACTION_DRAW_NEXT,
		w->tertris.nextThing.randomNumber,
		w->tertris.nextThing.randomRotation,
		w->tertris.nextThing.randomReflection,
		w->tertris.units, w->tertris.diagonal, w->tertris.mixed,
		w->tertris.next.bonus);
}

static void
drawStatus(TertrisWidget w)
{
	setGame(w, ACTION_STATUS);
}

void
redoNext(TertrisWidget w)
{
	int diagonal = (w->tertris.next.diagonal) ? 1 : 0;
	int turnStyle = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	int next_start, i;

	w->tertris.nextThing.units = w->tertris.next.units - MIN_UNITS +
		((w->tertris.nextThing.bonus) ? 1 : 0);
	next_start = tris[w->tertris.nextThing.units].mode
		[diagonal].number[(w->tertris.nextThing.bonus) ?
		NOMIX : ((w->tertris.next.mixed) ? 1 : 0)][turnStyle];
	if (!next_start) {
		DISPLAY_ERROR("Error, no polyiamonds for this combination");
	}
	next_start = w->tertris.nextThing.randomNumber % next_start;
	w->tertris.nextThing.colorNumber = next_start;

	if (w->tertris.next.mixed && !w->tertris.nextThing.bonus) {
		w->tertris.nextThing.units = 0;
		while (next_start >= tris[w->tertris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle]) {
			next_start -= tris[w->tertris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle];
			w->tertris.nextThing.units++;
		}
	}
	w->tertris.nextThing.polyNumber =
		tris[w->tertris.nextThing.units].mode
		[diagonal].start[next_start][turnStyle];
	if (tris[w->tertris.nextThing.units].mode[diagonal].turnStyle > NONE)
		for (i = 0; i < w->tertris.nextThing.randomRotation; i++)
			w->tertris.nextThing.polyNumber =
				tris[w->tertris.nextThing.units].polyiamond
				[w->tertris.nextThing.polyNumber].rotation;
	if (tris[w->tertris.nextThing.units].mode[diagonal].turnStyle == ALL)
		for (i = 0; i < w->tertris.nextThing.randomReflection; i++)
			w->tertris.nextThing.polyNumber =
				tris[w->tertris.nextThing.units].polyiamond
				[w->tertris.nextThing.polyNumber].reflection;
	w->tertris.nextThing.size = tris[w->tertris.nextThing.units].polyiamond
		[w->tertris.nextThing.polyNumber].size;
	/*nextThing.xpos = NRAND(w->tertris.columns -
		w->tertris.nextThing.size + 1); */
	w->tertris.nextThing.xpos = (w->tertris.columns -
		w->tertris.nextThing.size) / 2 - 2;
	w->tertris.nextThing.ypos =
		-tris[w->tertris.nextThing.units].polyiamond
		[w->tertris.nextThing.polyNumber].startHeight
		[(w->tertris.suddenAppear) ? JUMPIN : GRADUAL];
	if ((w->tertris.nextThing.xpos - w->tertris.nextThing.ypos +
			!w->tertris.invert + w->tertris.nextThing.size) % 2)
		w->tertris.nextThing.xpos++;
	w->tertris.nextThing.direction = (NRAND(2) % 2) ? 1 : -1;
}

static void
initFill(TertrisWidget w)
{
	int i, j, col, offset, fixedUnits;

	if (w->tertris.fixedFill) {
		/* This algorithm randomly picks fixedUnits per row
		 * (4 for DEFAULT_COLUMNS, ie. 10/2-1)
		 * This is more in keeping with the original */
		fixedUnits = w->tertris.columns / 2 - 1;
		for (j = 0; j < w->tertris.rows; j++) {
			offset = j * w->tertris.columns;
			for (i = 0; i < w->tertris.columns; i++) {
				w->tertris.field[offset + i].pmid = -1;
				w->tertris.field[offset + i].cid = 0;
			}
			if (j >= w->tertris.rows - w->tertris.fill)
				for (i = 0; i < fixedUnits; i++) {
					do
						col = NRAND(w->tertris.columns);
					while (w->tertris.field[offset + col].pmid != -1);
					w->tertris.field[offset + col].pmid = BITMAPS - 1;
					w->tertris.field[offset + col].cid = NRAND(w->tertris.numberPolyiamonds);
				}
		}
	} else {
		/* This algorithm randomly fills boxes to a certain level
		 * The problem with this is that it could potentially fill
		 * up a whole row */
		for (j = 0; j < w->tertris.rows; j++) {
			offset = j * w->tertris.columns;
			for (i = 0; i < w->tertris.columns; i++)
				if ((j >= w->tertris.rows - w->tertris.fill) &&
						(NRAND(2) == 0)) {
					w->tertris.field[offset + i].pmid = BITMAPS - 1;
					w->tertris.field[offset + i].cid = NRAND(w->tertris.numberPolyiamonds);
				} else {
					w->tertris.field[offset + i].pmid = -1;
					w->tertris.field[offset + i].cid = 0;
				}
		}
	}
}

static void
fillLines(TertrisWidget w)
{
	int i, j, k;
	int diagonal = (w->tertris.current.diagonal) ? 1 : 0;
	int mixed = (w->tertris.current.mixed) ? 1 : 0;
	int turnStyle = tris[w->tertris.current.units -
		MIN_UNITS].mode[diagonal].turnStyle;

	FLUSH(w);
	for (j = 0; j <= w->tertris.rows / 2; j++) {
		k = w->tertris.rows - j - 1;
		for (i = 0; i < w->tertris.columns; i++) {
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->tertris.current.units -
				MIN_UNITS].mode[diagonal].number[mixed][turnStyle]), i, j, False);
			drawUnitBorder(w, i, j);
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->tertris.current.units -
				MIN_UNITS].mode[diagonal].number[mixed][turnStyle]), i, k, False);
			drawUnitBorder(w, i, k);
		}
		FLUSH(w);
		Sleep(100);
	}
	Sleep(1000);
}

static Boolean
overlapping(TertrisWidget w)
{
	int i, j;
	int x = w->tertris.curThing.xpos, y = w->tertris.curThing.ypos;

	for (i = 0; i < HEX_COL(w->tertris.curThing.size); i++)
		for (j = 0; j < HEX_ROW(w->tertris.curThing.size); j++)
			if (tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape[j][i] != 0) {
				if ((y + j >= w->tertris.rows) ||
						(x + i < 0) ||
						(x + i >= w->tertris.columns))
					return True;
				if (w->tertris.suddenAppear) { /* creationalism */
/* This method does not allow a turning polyiamond to an area above screen */
					if ((y + j < 0) ||
							w->tertris.field[(y + j) * w->tertris.columns + x + i].pmid >= 0) {
						return True;
					}
				} else { /* evolution */
/* This method one can turn polyiamond to an area above of screen,
 * also part of the polyiamond may not be visible initially */
					if ((y + j >= 0) &&
							w->tertris.field[(y + j) * w->tertris.columns + x + i].pmid >= 0) {
						return True;
					}
				}
			}
	return False;
}

static Boolean
overlapped(TertrisWidget w, int dir, int fall)
{
	int i, j;
	int oldx = w->tertris.curThing.xpos - (!fall + 1) * dir;
	int oldy = w->tertris.curThing.ypos - fall;

	for (i = 0; i < HEX_COL(w->tertris.curThing.size); i++)
		for (j = 0; j < HEX_ROW(w->tertris.curThing.size); j++)
			if ((w->tertris.curThing.ypos + j >= 0) &&
					tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape[j][i] != 0) {
				if (((oldx + i + oldy + j + w->tertris.invert) & 1) == 1) {
					if ((oldy + j >= 0) &&
							(w->tertris.field[(oldy + j) * w->tertris.columns + oldx + i + dir].pmid >= 0))
						return True;
				} else {
					if (w->tertris.field[(oldy + j + fall) * w->tertris.columns + oldx + i + !fall * dir].pmid >= 0)
						return True;
				}
			}
	return False;
}

static Boolean
collision(TertrisWidget w, int dir)
{
	int i, j;
	int x = w->tertris.curThing.xpos, y = w->tertris.curThing.ypos;

	for (i = 0; i < HEX_COL(w->tertris.curThing.size); i++)
		for (j = 0; j < HEX_ROW(w->tertris.curThing.size); j++)
			if ((y + j >= -1) &&
					tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape[j][i] != 0) {
				if ((y + j >= w->tertris.rows - 1) ||
						(x + i + dir < 0) ||
						(x + i + dir >= w->tertris.columns) ||
						w->tertris.field[(y + j + 1) * w->tertris.columns + x + i + dir].pmid >= 0)
					return True;
				if (((x + i + y + j + w->tertris.invert) & 1) == 1) {
					if ((y + j >= 0) &&
							w->tertris.field[(y + j) * w->tertris.columns + x + i + dir].pmid >= 0)
						return True;
				} else {
					if (w->tertris.field[(y + j + 1) * w->tertris.columns + x + i].pmid >= 0)
						return True;
				}
			}
	return False;
}

static Boolean
atBottom(TertrisWidget w)
{
	if (collision(w, w->tertris.curThing.direction))
		return collision(w, -w->tertris.curThing.direction);
	return False;
}

static void
putTri(TertrisWidget w)
{
	int i, j;

	for (i = 0; i < HEX_COL(w->tertris.curThing.size); i++) {
		int xi = w->tertris.curThing.xpos + i;

		for (j = 0; j < HEX_ROW(w->tertris.curThing.size); j++) {
			int yj = w->tertris.curThing.ypos + j;

			if ((yj >= 0) &&
					tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape[j][i] != 0) {
				w->tertris.field[yj * w->tertris.columns + xi].pmid =
					(polyPixmap(i, j,
					(tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].shape),
					w->tertris.curThing.size)) % BITMAPS;
				w->tertris.field[yj * w->tertris.columns + xi].cid =
					w->tertris.curThing.colorNumber;
			}
		}
	}
}

static int
checkLines(TertrisWidget w)
{
	int *lSet, nSet = 0;
	int i, j, k, y, offset;

	if (!(lSet = (int *) malloc(sizeof (int) * w->tertris.rows))) {
		DISPLAY_ERROR("Not enough memory (lSet), exiting.");
	}
	/* Find filled rows */
	for (j = 0; j < w->tertris.rows; j++) {
		lSet[j] = 0;
		offset = j * w->tertris.columns;
		for (i = 0; i < w->tertris.columns; i++)
			if (w->tertris.field[offset + i].pmid >= 0)
				lSet[j]++;
		if (lSet[j] == w->tertris.columns)
			nSet++;
	}

	if (nSet == 0) {
		free(lSet);
		return 0;
	}
	for (k = 0; k < 2 * NUM_FLASHES; k++) {
		for (j = 0; j < w->tertris.rows; j++) {
			if (lSet[j] == w->tertris.columns) {
				for (i = 0; i < w->tertris.columns; i++) {
					xorUnit(w, i, j, k);
					drawUnitBorder(w, i, j);
				}
			}
		}
		FLUSH(w);
		Sleep(100);
	}
	/* Shrink */
	for (j = w->tertris.rows - 1; j >= 0; j--)
		if (lSet[j] == w->tertris.columns) {
			for (y = j; y > 0; y--)
				for (i = 0; i < w->tertris.columns; i++) {
					w->tertris.field[y * w->tertris.columns + i] =
						w->tertris.field[(y - 1) * w->tertris.columns + i];
					moveUnit(w, i, y);
				}
			for (i = 0; i < w->tertris.columns; i++) {
				w->tertris.field[i].pmid = -1;
				clearUnit(w, i, 0);
				drawUnitBorder(w, i, 0);
			}
			/* drawPartialField(w, 0, j + 1); */
			for (i = j; i > 0; i--)
				lSet[i] = lSet[i - 1];
			lSet[0] = 0;
			if (j > 0)
				for (i = 0; i < w->tertris.columns; i++) {
					int tmp = w->tertris.field[j * w->tertris.columns + i].pmid;

					if ((((i + j + ((w->tertris.invert) ? 0 : 1)) & 1) == 0) &&
							(tmp >= 0) && (tmp != CHECKDOWN(tmp))) {
						w->tertris.field[j * w->tertris.columns + i].pmid = CHECKDOWN(tmp);
						drawUnit(w, w->tertris.field[j * w->tertris.columns + i].pmid,
						w->tertris.field[j * w->tertris.columns + i].cid, i, j, False);
					}
				}
			FLUSH(w);
			j++;
			for (k = j; k < w->tertris.rows; k++) {
				offset = k * w->tertris.columns;
				for (i = 0; i < w->tertris.columns; i++) {
					int tmp = w->tertris.field[offset + i].pmid;

					if (tmp >= 0) {
						if (((i + k + ((w->tertris.invert) ? 0 : 1)) & 1) == 1)
							tmp = CHECKDOWN(tmp);
						if (((i + k + ((w->tertris.invert) ? 0 : 1)) & 1) == 0)
							tmp = CHECKUP(tmp);
						w->tertris.field[offset + i].pmid = SWAPUPDOWN(tmp);
					}
				}
			}
#ifdef USE_SOUND
			if (w->tertris.sound) {
				playSound((char *) BUMPSOUND);
			}
#endif
		}
	free(lSet);
	clearField(w);
	if ((nSet & 1) == 1) {
		w->tertris.invert = !w->tertris.invert;
		w->tertris.nextThing.xpos =
			(w->tertris.columns - w->tertris.nextThing.units) / 2;
		if ((w->tertris.nextThing.xpos - w->tertris.nextThing.ypos +
				((w->tertris.invert) ? 0 : 1) +
				w->tertris.nextThing.size) % 2 != 0)
			w->tertris.nextThing.xpos++;
	}
	drawField(w);
	if (w->tertris.grid)
		drawGrid(w);
	FLUSH(w);
	return nSet;
}


static void
newThing(TertrisWidget w)
{
	w->tertris.curThing = w->tertris.nextThing;
	/*w->tertris.current = w->tertris.next;*/
	w->tertris.current.units = w->tertris.next.units;
	w->tertris.current.diagonal = w->tertris.next.diagonal;
	w->tertris.current.mixed = w->tertris.next.mixed;
	w->tertris.current.bonus = w->tertris.next.bonus;
	w->tertris.nextThing.randomNumber = LRAND();
	w->tertris.nextThing.randomRotation = NRAND(MAX_SIDES);
	w->tertris.nextThing.randomReflection = NRAND(2);
	w->tertris.nextThing.bonus = (w->tertris.next.bonus) ? 1 : 0;
	redoNext(w);
}

static void
clearThings(TertrisWidget w, Boolean draw)
{
	if (draw)
		clearField(w);
	initFill(w);
	if (draw) {
		if (w->tertris.grid)
			drawGrid(w);
		drawField(w);
	}
	w->tertris.delay = delay[w->tertris.level - 1];
	w->tertris.score = w->tertris.fill * w->tertris.level * 10;
	w->tertris.started = True;
	w->tertris.paused = False;
	w->tertris.pauseStart = 0;
	w->tertris.completedRows = 0;
	w->tertris.newRows = 0;
	drawStatus(w);
}

static void
gameOver(TertrisWidget w, Boolean fast)
{
	if (!w->tertris.started)
		return;
	if (!fast)
		fillLines(w);
	setGame(w, ACTION_GAME_OVER);
	clearThings(w, True);
}

static void
tryMove(TertrisWidget w, move_t move)
{
	int diagonal = (w->tertris.current.diagonal) ? 1 : 0;
	Thing old, mid;
	int i;

	old = w->tertris.curThing;
	switch (move) {
	case FALL:
		w->tertris.curThing.ypos++;
		w->tertris.curThing.xpos += w->tertris.curThing.direction;
		if (overlapping(w) || overlapped(w, w->tertris.curThing.direction, True)) {
			w->tertris.curThing = old;
			w->tertris.curThing.direction = -w->tertris.curThing.direction;
			w->tertris.curThing.ypos++;
			w->tertris.curThing.xpos += w->tertris.curThing.direction;
		}
		if (overlapping(w) || overlapped(w, w->tertris.curThing.direction, True))
			w->tertris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	case DROP:
		do {
			do {
				mid = w->tertris.curThing;
				w->tertris.curThing.ypos++;
				w->tertris.curThing.xpos += w->tertris.curThing.direction;
				if (w->tertris.dropPaused == 0)
					w->tertris.score += w->tertris.level +
						!w->tertris.grid + w->tertris.fill;
			} while (!overlapping(w) && !overlapped(w, w->tertris.curThing.direction, True));
				w->tertris.curThing = mid;
				w->tertris.curThing.direction = -w->tertris.curThing.direction;
				w->tertris.curThing.ypos++;
				w->tertris.curThing.xpos += w->tertris.curThing.direction;
		} while (!overlapping(w) && !overlapped(w, w->tertris.curThing.direction, True));
		w->tertris.curThing = mid;
		drawThingDiff(w, old);
		break;
	case MOVE_LEFT:
		w->tertris.curThing.xpos -= 2;
		if (overlapping(w) || overlapped(w, -1, False))
			w->tertris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	case MOVE_RIGHT:
		w->tertris.curThing.xpos += 2;
		if (overlapping(w) || overlapped(w, 1, False))
			w->tertris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	case ROTATE:
		if (tris[w->tertris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle > NONE) {
			if (w->tertris.cw)
				for (i = 0; i < MAX_SIDES - 1; i++)
					w->tertris.curThing.polyNumber = tris[w->tertris.curThing.units].polyiamond
						[w->tertris.curThing.polyNumber].rotation;
			else /* ccw */
				w->tertris.curThing.polyNumber = tris[w->tertris.curThing.units].polyiamond
					[w->tertris.curThing.polyNumber].rotation;
			w->tertris.curThing.xpos = old.xpos;
			w->tertris.curThing.ypos = old.ypos;
		}
		if (overlapping(w))
			w->tertris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	case REFLECT: /* reflect on y axis */
		if (tris[w->tertris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle == ALL) {
			w->tertris.curThing.polyNumber =
				tris[w->tertris.curThing.units].polyiamond
				[w->tertris.curThing.polyNumber].reflection;
			w->tertris.curThing.xpos = old.xpos;
			w->tertris.curThing.ypos = old.ypos;
		}
		if (overlapping(w))
			w->tertris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	}
}

static Boolean
moveOne(TertrisWidget w, move_t move)
{
	int lines;

	if ((move == DROP) || ((move == FALL) && atBottom(w))) {
		tryMove(w, move);
		putTri(w);
		lines = checkLines(w);
		if (w->tertris.dropPaused != 0) {
			w->tertris.dropPaused = 0;
		} else {
			w->tertris.score += (!w->tertris.grid + w->tertris.fill +
				w->tertris.level + lines + w->tertris.suddenAppear) *
				(w->tertris.current.bonus ? 1 : 3) *
				(w->tertris.showNext ? 1 : 3) /
				(1 + w->tertris.current.mixed);
			w->tertris.score += (w->tertris.showNext ? 3 : 5) *
				w->tertris.level * w->tertris.level * lines * lines *
				(w->tertris.current.bonus ? 3 : 5) *
				w->tertris.startLevel *
				(w->tertris.current.units + w->tertris.current.diagonal) *
				(w->tertris.current.units + w->tertris.current.diagonal) /
				(16 * (1 + w->tertris.current.mixed));
		}
		w->tertris.completedRows += lines;
		w->tertris.newRows += lines;
		w->tertris.next.bonus = False;
		if ((w->tertris.newRows > THRESHOLD(w->tertris.level)) &&
				(w->tertris.level < MAX_LEVELS)) {
			w->tertris.level++;
			w->tertris.newRows = 0;
			if (LEGAL(w->tertris.next.units,
					w->tertris.next.diagonal,
					w->tertris.next.mixed,
					w->tertris.bonus))
				w->tertris.next.bonus = True;
		}
		drawStatus(w);
		newThing(w);
		drawThing(w);
		drawNext(w);
		FLUSH(w);
		if (overlapping(w))
			gameOver(w, False);
		return True;
	} else {
		if (w->tertris.dropPaused == 1)
			w->tertris.dropPaused = 2;
		tryMove(w, move);
		if ((w->tertris.newRows > THRESHOLD(w->tertris.level)) &&
				(w->tertris.level < MAX_LEVELS)) {
			w->tertris.next.bonus = False;
			w->tertris.level++;
			w->tertris.newRows = 0;
			if (LEGAL(w->tertris.next.units,
					w->tertris.next.diagonal,
					w->tertris.next.mixed,
					w->tertris.bonus))
				w->tertris.next.bonus = True;
			drawStatus(w);
			drawNext(w);
		}
		return False;
	}
}

#ifndef WINVER
static
#endif
void
falldownGame(TertrisWidget w)
{
	if (!w->tertris.paused) {
		w->tertris.gotNew = moveOne(w, FALL) || w->tertris.gotNew;
	} else {
		if (w->tertris.pauseStart >= 32) {
			clearThings(w, True);
		}
		if (!w->tertris.started)
			w->tertris.pauseStart++;
	}
}

#ifndef WINVER
static
#endif
void
moveGameInput(TertrisWidget w, int direction)
{
	if (!w->tertris.paused)
		w->tertris.gotNew = moveOne(w, (move_t) direction) ||
			w->tertris.gotNew;
}

static void
resetThings(TertrisWidget w)
{
	if (w->tertris.field)
		free(w->tertris.field);
	w->tertris.rows = (w->tertris.columns / 3) * 4;
	if (!(w->tertris.field = (Field *) malloc(sizeof(Field) *
			w->tertris.rows * w->tertris.columns))) {
		DISPLAY_ERROR("Not enough memory (field), exiting.");
	}
	w->tertris.started = False;
}

#if 0
static void
resizeThings(TertrisWidget w)
{
}
#endif

#ifdef WINVER
static void
setValuesGame(TertrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->tertris.level = GetPrivateProfileInt(SECTION,
		"level", DEFAULT_LEVELS, INIFILE);
	w->tertris.units = GetPrivateProfileInt(SECTION,
		"units", DEFAULT_UNITS, INIFILE);
	w->tertris.columns = GetPrivateProfileInt(SECTION,
		"columns", DEFAULT_COLUMNS, INIFILE);
	w->tertris.fill = GetPrivateProfileInt(SECTION,
		"fill", 0, INIFILE);
	w->tertris.fixedFill = (BOOL) GetPrivateProfileInt(SECTION,
		"fixedFill", TRUE, INIFILE);
	w->tertris.suddenAppear = (BOOL) GetPrivateProfileInt(SECTION,
		"suddenAppear", FALSE, INIFILE);
	w->tertris.diagonal = (BOOL) GetPrivateProfileInt(SECTION,
		"diagonal", DEFAULT_DIAGONAL, INIFILE);
	w->tertris.mixed = (BOOL) GetPrivateProfileInt(SECTION,
		"mixed", DEFAULT_MIXED, INIFILE);
	w->tertris.bonus = (BOOL) GetPrivateProfileInt(SECTION,
		"bonus", TRUE, INIFILE);
	w->tertris.cw = (BOOL) GetPrivateProfileInt(SECTION,
		"cw", TRUE, INIFILE);
	w->tertris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->tertris.showNext = (BOOL) GetPrivateProfileInt(SECTION,
		"showNext", FALSE, INIFILE);
	w->tertris.grid = (BOOL) GetPrivateProfileInt(SECTION,
		"grid", TRUE, INIFILE);
	w->tertris.sound = (BOOL) GetPrivateProfileInt(SECTION,
		"sound", FALSE, INIFILE);
	(void) GetPrivateProfileString(SECTION,
		"bumpSound", BUMPSOUND, szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->tertris.bumpSound, szBuf, 80);
	w->tertris.mono = (BOOL) GetPrivateProfileInt(SECTION,
		"mono", DEFAULT_MONO, INIFILE);
	w->tertris.reverse = (BOOL) GetPrivateProfileInt(SECTION,
		"reverseVideo", DEFAULT_REVERSE, INIFILE);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION,
		"background", "255 255 255", szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tertris.inverseGC[1] = RGB(color.red, color.green, color.blue);
	w->tertris.inverseGC[0] = RGB(brighter(color.red),
		brighter(color.green), brighter(color.blue));
	w->tertris.inverseGC[2] = RGB(darker(color.red),
		darker(color.green), darker(color.blue));
	w->tertris.inverseGC[3] = RGB(darker(darker(color.red)),
		darker(darker(color.green)), darker(darker(color.blue)));
	w->tertris.inverseGC[4] = RGB(darker(darker(darker(color.red))),
		darker(darker(darker(color.green))),
		darker(darker(darker(color.blue))));
	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION,
			buf, polyiamondColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->tertris.thingGC[thing][1] =
			RGB(color.red, color.green, color.blue);
		w->tertris.thingGC[thing][0] =
			RGB(brighter(color.red), brighter(color.green),
			brighter(color.blue));
		w->tertris.thingGC[thing][2] =
			RGB(darker(color.red), darker(color.green),
			darker(color.blue));
	}
	(void) GetPrivateProfileString(SECTION,
		"userName", "Guest", szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->tertris.userName, szBuf, 80);
	(void) GetPrivateProfileString(SECTION,
		"scoreFile", "", szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->tertris.scoreFile, szBuf, 80);
}

void
destroyGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	PostQuitMessage(0);
}

#else
static void
getColor(TertrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->tertris.reverse) {
		values.background = w->tertris.foreground;
	} else {
		values.background = w->tertris.background;
	}
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingGC[thing][1])
		XtReleaseGC((Widget) w, w->tertris.thingGC[thing][1]);
	w->tertris.thingGC[thing][1] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				brighter(w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingGC[thing][0])
		XtReleaseGC((Widget) w, w->tertris.thingGC[thing][0]);
	w->tertris.thingGC[thing][0] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				darker(w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingGC[thing][2])
		XtReleaseGC((Widget) w, w->tertris.thingGC[thing][2]);
	w->tertris.thingGC[thing][2] = XtGetGC((Widget) w, valueMask,
		&values);
}

static void
setAllColors(TertrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyiamond;

	valueMask = GCForeground | GCBackground;
	if (w->tertris.reverse) {
		values.background = w->tertris.background;
		values.foreground = w->tertris.foreground;
	} else {
		values.foreground = w->tertris.background;
		values.background = w->tertris.foreground;
	}
	if (w->tertris.inverseGC[1])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[1]);
	w->tertris.inverseGC[1] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = brighter(w, (w->tertris.reverse) ?
			w->tertris.foreground : w->tertris.background);
	}
	if (w->tertris.inverseGC[0])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[0]);
	w->tertris.inverseGC[0] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = darker(w, (w->tertris.reverse) ?
			w->tertris.foreground : w->tertris.background);
	}
	if (w->tertris.inverseGC[2])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[2]);
	w->tertris.inverseGC[2] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = darker(w, darker(w, (w->tertris.reverse) ?
			w->tertris.foreground : w->tertris.background));
	}
	if (w->tertris.inverseGC[3])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[3]);
	w->tertris.inverseGC[3] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = darker(w, darker(w, darker(w,
			(w->tertris.reverse) ? w->tertris.foreground :
			w->tertris.background)));
	}
	if (w->tertris.inverseGC[4])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[4]);
	w->tertris.inverseGC[4] = XtGetGC((Widget) w, valueMask, &values);
	for (polyiamond = 0; polyiamond < MAX_START_POLYIAMONDS; polyiamond++)
		getColor(w, polyiamond);
	valueMask = GCForeground | GCBackground | GCFunction;
	values.foreground = 0xFFFFFF;
	values.background = 0;
	values.function = GXxor;
	if (w->tertris.xorGC)
		XtReleaseGC((Widget) w, w->tertris.xorGC);
	w->tertris.xorGC = XtGetGC((Widget) w, valueMask, &values);
}

static Boolean
setValuesGame(Widget current, Widget request, Widget renew)
{
	TertrisWidget c = (TertrisWidget) current, w = (TertrisWidget) renew;
	Boolean redraw = False, setColors = False;

	if (w->tertris.act) {
		falldownGame(w);
		w->tertris.act = False;
	}
	if (w->tertris.level != c->tertris.level)
		w->tertris.delay = delay[w->tertris.level - 1];
	checkThings(w);
	if (w->tertris.reverse != c->tertris.reverse ||
			w->tertris.mono != c->tertris.mono) {
		setAllColors(w);
		redraw = True;
	} else if (w->tertris.background != c->tertris.background ||
			w->tertris.foreground != c->tertris.foreground ||
			setColors) {
		setAllColors(w);
		redraw = True;
	} else if (w->tertris.plainPiece != c->tertris.plainPiece) {
		redraw = True;
	} else if (w->tertris.suddenAppear != c->tertris.suddenAppear) {
		redoNext(w);
	}
	if (w->tertris.columns != c->tertris.columns) {
		w->tertris.paused = True;
		w->tertris.started = False;
		eraseFrame(w);
		sizeGame(w);
#if 0
		resizeThings(w);
#endif
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->tertris.focus);
		if (w->tertris.grid)
			drawGrid(w);
		redraw = True;
	} else if (w->tertris.fill != c->tertris.fill) {
		clearThings(w, True);
		redraw = True;
	} else if (w->tertris.units != c->tertris.units) {
		w->tertris.next.units = w->tertris.units;
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tertris.diagonal != c->tertris.diagonal) {
		w->tertris.next.diagonal = w->tertris.diagonal;
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tertris.mixed != c->tertris.mixed) {
		w->tertris.next.mixed = w->tertris.mixed;
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tertris.bonus != c->tertris.bonus) {
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tertris.grid != c->tertris.grid) {
		eraseFrame(w);
		drawFrame(w, w->tertris.focus);
		if (w->tertris.grid)
			drawGrid(w);
		drawField(w);
	}
	if (w->tertris.menu != -1) {
		switch (w->tertris.menu) {
		case 1:
			w->tertris.paused = True;
			break;
		case 2:
			w->tertris.menu = -1;
			clearThings(w, True);
			redraw = True;
			break;
		case 4:
			w->tertris.menu = -1;
			/*ConfirmThings(w);*/
			redraw = True;
			break;
		default:
			w->tertris.menu = -1;
			break;
		}
	}
	return (redraw);
}

static void
quitGame(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	gameOver(w, True);
	XtCloseDisplay(display);
	exit(0);
}

static void
destroyGame(Widget old)
{
	TertrisWidget w = (TertrisWidget) old;
	int thing, i;

#if defined( USE_SOUND ) && defined( USE_ESOUND )
	(void) shutdown_sound();
#endif
	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			XtReleaseGC(old, w->tertris.thingGC[thing][i]);
	}
	for (i = 0; i < BG_SHADES; i++)
		XtReleaseGC(old, w->tertris.inverseGC[i]);
	XtReleaseGC(old, w->tertris.xorGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->tertris.select);
}
#endif

#ifndef WINVER
static
#endif
void
resizeGame(TertrisWidget w)
{
	int i, j;
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif

	w->tertris.triHeight = MIN((w->core.height - 5) / w->tertris.rows,
		((w->core.width - 2) * 2 * SQRT3_2 / (w->tertris.columns + 1)));
	w->tertris.triWidth = (int) (w->tertris.triHeight / SQRT3_2);
	w->tertris.viewWidth = w->tertris.triWidth *
		(w->tertris.columns / 2 + 1) + 1;
	w->tertris.viewHeight = w->tertris.triHeight * w->tertris.rows + 1;
	w->tertris.offset.x = ((int) w->core.width -
		w->tertris.viewWidth) / 2;
	w->tertris.offset.y = ((int) w->core.height -
		w->tertris.viewHeight) / 2;

	for (j = 0; j < MAX_ORIENT; j++) {
		for (i = 0; i < 3; i++) {
			triangle_list[j][i].x =
				(w->tertris.triWidth - 1) / 2 *
				triangle_unit[j][i].x;
			triangle_list[j][i].y =
				(w->tertris.triHeight - 1) *
				triangle_unit[j][i].y;
			triangle_border[j][i].x =
				w->tertris.triWidth / 2 * triangle_unit[j][i].x;
			triangle_border[j][i].y =
				w->tertris.triHeight * triangle_unit[j][i].y;
		}
		triangle_list[j][3].x = triangle_list[j][1].x;
		triangle_list[j][3].y = -triangle_list[j][1].y;
		triangle_border[j][3].x = triangle_border[j][1].x;
		triangle_border[j][3].y = -triangle_border[j][1].y;
	}
}

#ifndef WINVER
static
#endif
void
sizeGame(TertrisWidget w)
{
	resetThings(w);
	resizeGame(w);
}

#ifndef WINVER
static
#endif
void
initializeGame(
#ifdef WINVER
TertrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle;
#ifndef WINVER
	int thing, i;
	TertrisWidget w = (TertrisWidget) renew;
#endif

#ifdef WINVER
	setValuesGame(w);
#else
	w->tertris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->tertris.mono);
	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			w->tertris.thingGC[thing][i] = NULL;
	}
	for (i = 0; i < BG_SHADES; i++)
		w->tertris.inverseGC[i] = NULL;
	w->tertris.xorGC = NULL;
#endif
	w->tertris.field = NULL;
	w->tertris.started = False;
	w->tertris.quit = False;
	w->tertris.completedRows = 0;
	w->tertris.newRows = 0;
	w->tertris.fall = False;
	w->tertris.invert = False;
	w->tertris.focus = False;
	w->tertris.next.units = w->tertris.units;
	w->tertris.next.diagonal = w->tertris.diagonal;
	w->tertris.next.mixed = w->tertris.mixed;
	w->tertris.next.bonus = False;
	w->tertris.startLevel = w->tertris.level;
	w->tertris.nextThing.units = 0;
	w->tertris.nextThing.polyNumber = 0;
	w->tertris.nextThing.xpos = 0;
	w->tertris.nextThing.ypos = 0;
	w->tertris.nextThing.size = 0;
	w->tertris.nextThing.colorNumber = 0;
	w->tertris.nextThing.direction = 0;
	w->tertris.nextThing.bonus = 0;
	w->tertris.nextThing.randomRotation = 0;
	w->tertris.nextThing.randomReflection = 0;
	w->tertris.nextThing.randomNumber = 0;
	w->tertris.gotNew = False;
	w->tertris.dropPaused = 0;
	checkThings(w);
	sizeGame(w);
#ifdef WINVER
	brush = CreateSolidBrush(w->tertris.inverseGC[2]);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	setAllColors(w);
#endif
	diagonal = (w->tertris.next.diagonal) ? 1 : 0;
	mixed = (w->tertris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "Tertris: corrupted polyiamond file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->tertris.numberPolyiamonds = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->tertris.numberPolyiamonds <= 0 ||
			w->tertris.numberPolyiamonds > MAX_START_POLYIAMONDS) {
		(void) fprintf(stderr, "Tertris: corrupted polyiamond file\n");
		(void) fprintf(stderr, "\tnumberPolyiamonds = %d.\n",
			w->tertris.numberPolyiamonds);
		exit(1);
	}
	clearThings(w, False);
	newThing(w);
	newThing(w);
#ifdef USE_SOUND
#ifdef USE_NAS
	dsp = XtDisplay(w);
#endif
#ifdef USE_ESOUND
	(void) init_sound();
#endif
#endif
}

#ifndef WINVER
static
#endif
void
exposeGame(
#ifdef WINVER
TertrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	TertrisWidget w = (TertrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	eraseFrame(w);
#if 0
	resizeThings(w);
#endif
	drawFrame(w, w->tertris.focus);
	if (w->tertris.grid)
		drawGrid(w);
	drawField(w);
	drawThing(w);
	drawNext(w);
}

#ifndef WINVER
static
#endif
void
hideGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->tertris.sound)
		w->tertris.sound = False;
	w->tertris.paused = True;
	w->tertris.dropPaused = 2;
	setGame(w, ACTION_HIDE);
}

#ifndef WINVER
static
#endif
void
clearGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	clearThings(w, True);
}

#ifndef WINVER
static
#endif
void
enterGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->tertris.focus = True;
	w->tertris.paused = False;
	drawFrame(w, w->tertris.focus);
}

#ifndef WINVER
static
#endif
void
leaveGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	/* if (w->tertris.firstFall) */
	w->tertris.focus = False;
	w->tertris.paused = True;
	if (w->tertris.dropPaused == 0)
		w->tertris.dropPaused = 1;
	else
		w->tertris.dropPaused = 2;
	drawFrame(w, w->tertris.focus);
}

#ifndef WINVER
static void
mouseGame(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	if (w->tertris.dropPaused == 1) /* clicking causes enter/exit events */
		w->tertris.dropPaused = 0;
}

static void
reflectGame(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, REFLECT);
}

static void
rotateGame(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, ROTATE);
}

static void
moveGameLeft(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, MOVE_LEFT);
}

static void
moveGameRight(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, MOVE_RIGHT);
}

static void
dropGame(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, DROP);
}

#endif

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameLevel(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused && w->tertris.level < MAX_LEVELS) {
		w->tertris.level++;
		w->tertris.delay = delay[w->tertris.level - 1];
		w->tertris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#ifdef WINVER
		return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameLevel(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
#ifdef CHEAT
	if (!w->tertris.paused && w->tertris.level > MIN_LEVELS) {
		w->tertris.level--;
		w->tertris.delay = delay[w->tertris.level - 1];
		w->tertris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#endif
#ifdef WINVER
		return False;
#endif
}

#ifndef WINVER
static
#endif
void
pauseGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->tertris.sound)
		w->tertris.sound = False;
	w->tertris.paused = !w->tertris.paused;
	w->tertris.dropPaused = 2;
#if 0
	if (w->tertris.paused) {
		banner(MSG_PAUSED);
		(void) XIconifyWindow(display, mainWin, screen_num);
	} else {
		drawNext(w);
	}
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameUnits(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused && LEGAL(w->tertris.next.units + 1,
			w->tertris.next.diagonal, w->tertris.next.mixed,
			False)) {
		setGame(w, ACTION_INC_UNITS);
#ifdef WINVER
		w->tertris.next.units = w->tertris.units;
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameUnits(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused && LEGAL(w->tertris.next.units - 1,
			w->tertris.next.diagonal, w->tertris.next.mixed,
			False)) {
		setGame(w, ACTION_DEC_UNITS);
#ifdef WINVER
		w->tertris.next.units = w->tertris.units;
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameColumns(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused && w->tertris.columns < MAX_COLUMNS - 1) {
#ifdef WINVER
		w->tertris.paused = True;
		w->tertris.started = False;
		eraseFrame(w);
		w->tertris.columns += 2;
		sizeGame(w);
#if 0
		resizeThings(w);
#endif
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->tertris.focus);
		if (w->tertris.grid)
			drawGrid(w);
		return True;
#else
		setGame(w, ACTION_INC_COLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameColumns(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused && w->tertris.columns > MIN_COLUMNS + 1) {
#ifdef WINVER
		w->tertris.paused = True;
		w->tertris.started = False;
		eraseFrame(w);
		w->tertris.columns -= 2;
		sizeGame(w);
#if 0
		resizeThings(w);
#endif
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->tertris.focus);
		if (w->tertris.grid)
			drawGrid(w);
		return True;
#else
		setGame(w, ACTION_DEC_COLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
fixedfillGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_FIXED_FILL);
}

#ifndef WINVER
static
#endif
void
suddenAppearGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_SUDDEN_APPEAR);
#ifdef WINVER
	redoNext(w);
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
diagonalGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused && LEGAL(w->tertris.next.units,
			!w->tertris.next.diagonal, w->tertris.next.mixed,
			False)) {
		setGame(w, ACTION_DIAGONAL);
#ifdef WINVER
		w->tertris.next.diagonal = w->tertris.diagonal;
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
mixedGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused && LEGAL(w->tertris.next.units,
			w->tertris.next.diagonal, !w->tertris.next.mixed,
			False)) {
		setGame(w, ACTION_MIXED);
#ifdef WINVER
		w->tertris.next.mixed = w->tertris.mixed;
		w->tertris.next.bonus = False;
		w->tertris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
bonusGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_BONUS);
#ifdef WINVER
	w->tertris.next.bonus = False;
	w->tertris.newRows = 0;
	redoNext(w);
	drawNext(w);
#endif
}

#ifndef WINVER
static
#endif
void
cwGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_CW);
}

#ifndef WINVER
static
#endif
void
plainPieceGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_PLAIN_PIECE);
#ifdef WINVER
	drawField(w);
#endif
}

#ifndef WINVER
static
#endif
void
showNextGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused) {
		setGame(w, ACTION_SHOW_NEXT);
	}
}

#ifndef WINVER
static
#endif
void
gridGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused) {
		setGame(w, ACTION_GRID);
#ifdef WINVER
		eraseFrame(w);
		drawFrame(w, w->tertris.focus);
		if (w->tertris.grid)
			drawGrid(w);
		drawField(w);
#endif
	}
}

#ifndef WINVER
static
#endif
void
soundGame(TertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_SOUND);
}

#ifndef WINVER
static
#endif
void
fillGame(TertrisWidget w, int n)
{
	setGame(w, n + ACTION_FILL0);
#ifdef WINVER
	clearThings(w, True);
#endif
}

#ifndef WINVER
static void
fill0Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 0);
}

static void
fill1Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 1);
}

static void
fill2Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 2);
}

static void
fill3Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 3);
}

static void
fill4Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 4);
}

static void
fill5Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 5);
}

static void
fill6Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 6);
}

static void
fill7Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 7);
}

static void
fill8Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 8);
}

static void
fill9Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 9);
}

static void
fill10Game(TertrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 10);
}
#endif
